from trilogy.core.enums import (
    SourceType,
)
from trilogy.core.models.build import BuildGrain
from trilogy.core.models.environment import Environment
from trilogy.core.models.execute import QueryDatasource
from trilogy.core.processing.concept_strategies_v3 import History, search_concepts
from trilogy.core.query_processor import get_query_datasources, process_query
from trilogy.core.statements.author import SelectStatement


def test_direct_select(test_environment, test_environment_graph):
    history = History(base_environment=test_environment)
    test_environment = test_environment.materialize_for_select()
    product = test_environment.concepts["product_id"]
    #        concept, grain: Grain, environment: Environment, g: ReferenceGraph, query_graph: ReferenceGraph
    # test_environment = test_environment.materialize_for_select({})
    datasource = search_concepts(
        [product] + [test_environment.concepts[c] for c in product.grain.components],
        history=history,
        environment=test_environment,
        g=test_environment_graph,
        depth=0,
    ).resolve()

    assert isinstance(datasource, QueryDatasource)
    assert set([datasource.name for datasource in datasource.datasources]) == {
        "products"
    }


def test_get_datasource_from_window_function(
    test_environment: Environment, test_environment_graph
):
    history = History(base_environment=test_environment)
    test_environment = test_environment.materialize_for_select()
    product_rank = test_environment.concepts["product_revenue_rank"]
    #        concept, grain: Grain, environment: Environment, g: ReferenceGraph, query_graph: ReferenceGraph
    # assert product_rank.grain.components[0] == test_environment.concepts['name']
    node = search_concepts(
        [product_rank]
        + [test_environment.concepts[c] for c in product_rank.grain.components],
        history=history,
        environment=test_environment,
        g=test_environment_graph,
        depth=0,
    )

    # raise SyntaxError(product_rank.grain)
    datasource = node.resolve()
    assert product_rank in datasource.output_concepts
    # assert datasource.grain == product_rank.grain
    assert isinstance(datasource, QueryDatasource)
    assert datasource.grain.components == product_rank.grain.components

    product_rank_by_category = test_environment.concepts[
        "product_revenue_rank_by_category"
    ]
    #        concept, grain: Grain, environment: Environment, g: ReferenceGraph, query_graph: ReferenceGraph
    datasource = search_concepts(
        [product_rank_by_category]
        + [
            test_environment.concepts[c]
            for c in product_rank_by_category.grain.components
        ],
        history=history,
        environment=test_environment,
        g=test_environment_graph,
        depth=0,
    ).resolve()
    assert product_rank_by_category in datasource.output_concepts
    # assert datasource.grain == product_rank_by_category.grain

    assert isinstance(datasource, QueryDatasource)


def test_get_datasource_for_filter(
    test_environment: Environment, test_environment_graph
):
    history = History(base_environment=test_environment)
    test_environment = test_environment.materialize_for_select()
    hi_rev_product = test_environment.concepts["products_with_revenue_over_50"]
    #        concept, grain: Grain, environment: Environment, g: ReferenceGraph, query_graph: ReferenceGraph

    assert {n.name for n in hi_rev_product.sources} == {
        "total_revenue",
        "revenue",
        "product_id",
    }
    datasource = search_concepts(
        [hi_rev_product]
        + [test_environment.concepts[c] for c in hi_rev_product.grain.components],
        history=history,
        environment=test_environment,
        g=test_environment_graph,
        depth=0,
    )
    datasource = datasource.resolve()
    assert isinstance(datasource, QueryDatasource)
    assert hi_rev_product in datasource.output_concepts


def test_select_output(test_environment, test_environment_graph):
    history = History(base_environment=test_environment)
    test_environment = test_environment.materialize_for_select()
    product = test_environment.concepts["product_id"]
    #        concept, grain: Grain, environment: Environment, g: ReferenceGraph, query_graph: ReferenceGraph

    datasource = search_concepts(
        [product] + [test_environment.concepts[c] for c in product.grain.components],
        history=history,
        environment=test_environment,
        g=test_environment_graph,
        depth=0,
    ).resolve()

    assert isinstance(datasource, QueryDatasource)
    assert set([datasource.name for datasource in datasource.datasources]) == {
        "products"
    }


def test_basic_aggregate(test_environment: Environment, test_environment_graph):
    history = History(base_environment=test_environment)
    test_environment = test_environment.materialize_for_select()
    product = test_environment.concepts["product_id"]
    total_revenue = test_environment.concepts["total_revenue"]
    #        concept, grain: Grain, environment: Environment, g: ReferenceGraph, query_graph: ReferenceGraph
    datasource = search_concepts(
        [total_revenue.with_grain(product), product],
        history=history,
        environment=test_environment,
        g=test_environment_graph,
        depth=0,
    )
    datasource = datasource.resolve()
    assert isinstance(datasource, QueryDatasource)
    assert datasource.grain == BuildGrain(components={product.address})


def test_join_aggregate(test_environment: Environment, test_environment_graph):
    history = History(base_environment=test_environment)
    test_environment = test_environment.materialize_for_select()
    category_id = test_environment.concepts["category_id"]
    total_revenue = test_environment.concepts["total_revenue"]
    #        concept, grain: Grain, environment: Environment, g: ReferenceGraph, query_graph: ReferenceGraph
    datasource = search_concepts(
        [
            total_revenue.with_grain(BuildGrain(components={"local.category_id"})),
            category_id,
        ],
        history=history,
        environment=test_environment,
        g=test_environment_graph,
        depth=0,
    ).resolve()
    assert isinstance(datasource, QueryDatasource)
    assert datasource.source_type == SourceType.GROUP
    assert len(set([datasource.name for datasource in datasource.datasources])) == 1
    assert datasource.grain.components == {"local.category_id"}


def test_query_aggregation(test_environment, test_environment_graph):
    select = SelectStatement(selection=[test_environment.concepts["total_revenue"]])
    datasource = get_query_datasources(environment=test_environment, statement=select)

    assert {datasource.identifier} == {
        "revenue_at_local_order_id_grouped_by__at_abstract"
    }
    check = datasource
    assert len(check.input_concepts) == 2
    assert check.input_concepts[0].name == "revenue"
    assert len(check.output_concepts) == 1
    assert check.output_concepts[0].name == "total_revenue"


def test_query_datasources(test_environment, test_environment_graph):
    select = SelectStatement(
        selection=[
            test_environment.concepts["category_id"],
            test_environment.concepts["category_name"],
            test_environment.concepts["total_revenue"],
        ]
    )
    get_query_datasources(environment=test_environment, statement=select)


def test_full_query(test_environment, test_environment_graph):
    select = SelectStatement(
        selection=[
            test_environment.concepts["category_id"],
            test_environment.concepts["category_name"],
            test_environment.concepts["total_revenue"],
        ]
    )

    processed = process_query(statement=select, environment=test_environment)

    assert set(c.name for c in processed.output_columns) == {
        "category_id",
        "category_name",
        "total_revenue",
    }



from typing import TYPE_CHECKING, Any

from sqlalchemy import text

from trilogy.core.models.environment import Environment
from trilogy.dialect.duckdb import DuckDBDialect
from trilogy.engine import ExecutionEngine

if TYPE_CHECKING:
    try:
        from pandas import DataFrame
    except ImportError:
        DataFrame = Any


class DataframeDialect(DuckDBDialect):
    pass


class DataframeConnectionWrapper(ExecutionEngine):
    def __init__(self, engine: ExecutionEngine, dataframes: dict[str, "DataFrame"]):
        self.engine = engine
        self.dataframes = dataframes
        self.connection = None

    def setup(self, env: Environment, connection):
        self._register_dataframes(env, connection)

    def _register_dataframes(self, env: Environment, connection):
        for ds in env.datasources.values():
            if ds.safe_address in self.dataframes:
                connection.execute(
                    text("register(:name, :df)"),
                    {"name": ds.safe_address, "df": self.dataframes[ds.safe_address]},
                )
            else:
                raise ValueError(
                    f"Dataframe {ds.safe_address} not found in dataframes on connection config, have {self.dataframes.keys()}"
                )
        pass

    def add_dataframe(self, name: str, df: "DataFrame", connection, env: Environment):
        self.dataframes[name] = df
        self._register_dataframes(env, connection)

    def connect(self) -> Any:
        return self.engine.connect()

    def dispose(self, close=True):
        return super().dispose(close)



from typing import Optional

from trilogy.core.models.environment import Environment
from trilogy.parsing.parse_engine import parse_text


def parse(
    input: str, environment: Optional[Environment] = None
) -> tuple[Environment, list]:
    return parse_text(input, environment=environment)



from typing import Any, Callable, Mapping

from jinja2 import Template

from trilogy.core.enums import DatePart, FunctionType, WindowType
from trilogy.dialect.base import BaseDialect


def date_diff(first: str, second: str, grain: DatePart) -> str:
    grain = DatePart(grain)
    if grain == DatePart.YEAR:
        return f"date_part('year', {second}) - date_part('year', {first})"
    elif grain == DatePart.MONTH:
        return f"12 * {date_diff(first, second, DatePart.YEAR)} + date_part('month', {second}) - date_part('month', {first})"
    elif grain == DatePart.DAY:
        return f"date_part('day', {second} - {first})"
    elif grain == DatePart.HOUR:
        return f"{date_diff(first, second, DatePart.DAY)} *24 + date_part('hour', {second} - {first})"
    elif grain == DatePart.MINUTE:
        return f"{date_diff(first, second, DatePart.HOUR)} *60 + date_part('minute', {second} - {first})"
    elif grain == DatePart.SECOND:
        return f"{date_diff(first, second, DatePart.MINUTE)} *60 + date_part('second', {second} - {first})"
    else:
        raise NotImplementedError(f"Date diff not implemented for grain {grain}")


WINDOW_FUNCTION_MAP: Mapping[WindowType, Callable[[Any, Any, Any], str]] = {}

FUNCTION_MAP = {
    FunctionType.SPLIT: lambda x, types: f"string_to_array({x[0]}, {x[1]})",
    FunctionType.DATE_TRUNCATE: lambda x, types: f"date_trunc('{x[1]}', {x[0]})",
    FunctionType.DATE_ADD: lambda x, types: f"({x[0]} + INTERVAL '{x[2]} {x[1]}')",
    FunctionType.DATE_PART: lambda x, types: f"date_part('{x[1]}', {x[0]})",
    FunctionType.DATE_DIFF: lambda x, types: date_diff(x[0], x[1], x[2]),
    FunctionType.IS_NULL: lambda x, types: f"{x[0]} IS NULL",
}

FUNCTION_GRAIN_MATCH_MAP = {
    **FUNCTION_MAP,
    FunctionType.COUNT_DISTINCT: lambda args, types: f"CASE WHEN{args[0]} IS NOT NULL THEN 1 ELSE 0 END",
    FunctionType.COUNT: lambda args, types: f"CASE WHEN {args[0]} IS NOT NULL THEN 1 ELSE 0 END",
    FunctionType.SUM: lambda args, types: f"{args[0]}",
    FunctionType.AVG: lambda args, types: f"{args[0]}",
}

PG_SQL_TEMPLATE = Template(
    """{%- if output %}
DROP TABLE IF EXISTS {{ output.address.location }};
CREATE TABLE {{ output.address.location }} AS
{% endif %}{%- if ctes %}
WITH {% for cte in ctes %}
{{cte.name}} as ({{cte.statement}}){% if not loop.last %},{% endif %}{% endfor %}{% endif %}
{%- if full_select -%}
{{full_select}}
{%- else -%}
SELECT
{%- for select in select_columns %}
    {{ select }}{% if not loop.last %},{% endif %}{% endfor %}
{% if base %}FROM
    {{ base }}{% endif %}{% if joins %}
{% for join in joins %}
{{ join }}
{% endfor %}{% endif %}
{% if where %}WHERE
    {{ where }}
{% endif %}
{%- if group_by %}GROUP BY {% for group in group_by %}
    {{group}}{% if not loop.last %},{% endif %}{% endfor %}{% endif %}{% if having %}
HAVING
\t{{ having }}{% endif %}
{%- if order_by %}
ORDER BY {% for order in order_by %}
    {{ order }}{% if not loop.last %},{% endif %}
{% endfor %}{% endif %}
{%- if limit is not none %}
LIMIT {{ limit }}{% endif %}{% endif %}
"""
)

MAX_IDENTIFIER_LENGTH = 50


class PostgresDialect(BaseDialect):
    WINDOW_FUNCTION_MAP = {**BaseDialect.WINDOW_FUNCTION_MAP, **WINDOW_FUNCTION_MAP}
    FUNCTION_MAP = {**BaseDialect.FUNCTION_MAP, **FUNCTION_MAP}
    FUNCTION_GRAIN_MATCH_MAP = {
        **BaseDialect.FUNCTION_GRAIN_MATCH_MAP,
        **FUNCTION_GRAIN_MATCH_MAP,
    }
    QUOTE_CHARACTER = '"'
    SQL_TEMPLATE = PG_SQL_TEMPLATE



from pathlib import Path

from trilogy import Dialects, Environment
from trilogy.core.enums import Derivation, Purpose
from trilogy.hooks import DebuggingHook


def test_user_function_def():
    x = Dialects.DUCK_DB.default_executor()

    results = x.execute_query(
        """
def percent_ratio(a, b, digits=3) -> round(a::float / b * 100, digits);

                 
select @percent_ratio(10, 100) as ratio;
                 
                 """
    )

    assert results.fetchall()[0].ratio == 10.0


def test_user_function_def_with_default():
    x = Dialects.DUCK_DB.default_executor()
    DebuggingHook()
    results = x.execute_query(
        """
def percent_ratio(a, b, digits=3) -> round(a / b * 100, digits);

select @percent_ratio(21, 1000) as ratio,
         @percent_ratio(21, 1000, 0) as ratio_two;
                  
                  """
    )

    results = results.fetchall()
    assert results[0].ratio == 2.1
    assert results[0].ratio_two == 2


def test_user_function_aggregate():
    x = Dialects.DUCK_DB.default_executor()

    results = x.execute_query(
        """
key x int;
property x.price float;

datasource raw_data (
x: x,
price: price
)
grain (x)
query '''
select 1 as x, 2.0 as price
union all
select 2 as x, 3.0 as price
union all
select 10 as x, 5.0 as price
''';

def sum_times(a)-> a * sum(x);

                 
select @sum_times(10) as total;
                 
                 """
    )

    assert results.fetchall()[0].total == 130


def test_user_function_nested():
    x = Dialects.DUCK_DB.default_executor()

    results = x.execute_query(
        """
key x int;
property x.price float;

datasource raw_data (
x: x,
price: price
)
grain (x)
query '''
select 1 as x, 2.0 as price
union all
select 2 as x, 3.0 as price
union all
select 10 as x, 5.0 as price
''';

def sum_times(a)-> a * sum(x + price);

                 
select @sum_times(10) as total;
                 
                 """
    )

    assert results.fetchall()[0].total == 230


def test_user_function_nested_rowset():
    x = Dialects.DUCK_DB.default_executor()

    results = x.execute_query(
        """
key x int;
property x.price float;

datasource raw_data (
x: x,
price: price
)
grain (x)
query '''
select 1 as x, 2.0 as price
union all
select 2 as x, 3.0 as price
union all
select 10 as x, 5.0 as price
''';

def sum_times(a)-> a * sum(x + price);

with rowset as  
select @sum_times(10) as total;

select rowset.total;
                 
                 """
    )

    assert results.fetchall()[0].rowset_total == 230


def test_user_function_case():
    x = Dialects.DUCK_DB.default_executor()

    results = x.execute_query(
        """
key x int;
key y int;
property x.price float;

datasource raw_data (
x: x,
price: price
)
grain (x)
query '''
select 1 as x, 2.0 as price
union all
select 2 as x, 3.0 as price
union all
select 10 as x, 5.0 as price
''';

datasource join_x_y (
x: x,
y: y)
grain (x, y)
query '''
select 1 as x, 1 as y
union all
select 2 as x, 1 as y
union all
select 10 as x, 2 as y
''';

def weekday_sales(weekday) ->  
    SUM(CASE WHEN 10 = weekday THEN x ELSE 0 END) + 
    SUM(CASE WHEN 10 = weekday THEN price ELSE 0.0 END)
;


select 
    y,
    @weekday_sales(10) -> test
order by y asc;
                
"""
    )
    results = results.fetchall()
    assert results[0].test == 8
    assert results[1].test == 15


def test_parsing():
    x = Dialects.DUCK_DB.default_executor()
    x.execute_query(
        """
key x int;
property x.price float;

datasource raw_data (
x: x,
price: price
)
grain (x)
query '''
select 1 as x, 2.0 as price
union all
select 2 as x, 3.0 as price
union all
select 10 as x, 5.0 as price
''';


auto test <-SUM(CASE WHEN 10 = weekday THEN x ELSE 0 END) + 
    SUM(CASE WHEN 10 = weekday THEN price ELSE 0.0 END);



                 
"""
    )
    test = x.environment.concepts["test"]
    assert test.keys == set()
    assert test.purpose == Purpose.METRIC
    assert test.derivation == Derivation.BASIC


def test_user_function_import():
    env = Environment(working_path=Path(__file__).parent)
    x = Dialects.DUCK_DB.default_executor(environment=env)
    from trilogy.hooks import DebuggingHook

    DebuggingHook()
    results = x.execute_query(
        """
import test_env_functions as test_env_functions;

key x int;

merge test_env_functions.quad_test into x;

select 
    x as quad_test,
    @test_env_functions.quadratic(2, 3, 4) as quad_two;

"""
    )
    results = results.fetchall()
    assert results[0].quad_test == 16.414213562373096


def test_user_function_nesting():
    x = Dialects.DUCK_DB.default_executor()

    results = x.execute_query(
        """
key x int;
key y int;
property x.price float;

datasource raw_data (
x: x,
price: price
)
grain (x)
query '''
select 1 as x, 2.0 as price
union all
select 2 as x, 3.0 as price
union all
select 10 as x, 5.0 as price
''';

datasource join_x_y (
x: x,
y: y)
grain (x, y)
query '''
select 1 as x, 1 as y
union all
select 2 as x, 1 as y
union all
select 10 as x, 2 as y
''';

def weekday_sales(weekday) ->  
    SUM(CASE WHEN 10 = weekday THEN x ELSE 0 END) + 
    SUM(CASE WHEN 10 = weekday THEN price ELSE 0.0 END)
;

def plus_two(a) -> a + 2;

# auto random_no_f <- SUM(CASE WHEN 10 = weekday THEN x ELSE 0 END) + SUM(CASE WHEN 10 = weekday THEN price ELSE 0.0 END) +2;
auto random_one_f <- @weekday_sales(10) +2;
auto random <- @plus_two(@weekday_sales(10));
                
"""
    )
    # assert x.environment.concepts['random_no_f'].purpose == Purpose.METRIC, x.environment.concepts['random']
    assert (
        x.environment.concepts["random_one_f"].purpose == Purpose.METRIC
    ), x.environment.concepts["random"]
    assert (
        x.environment.concepts["random"].purpose == Purpose.METRIC
    ), x.environment.concepts["random"]

    results = x.execute_query(
        """select 
        y,
        @plus_two(@weekday_sales(10)) -> test2
    order by y asc;"""
    )
    results = results.fetchall()
    assert results[0].test2 == 10
    assert results[1].test2 == 17


def test_user_function_aggregate_two():
    x = Dialects.DUCK_DB.default_executor()

    results = x.execute_query(
        """import std.geography;
import std.display;

key state string::us_state;
key year int;
property state.val string;

datasource states
(
    state,
    year,
    val
)
grain (state, year)
query '''
select 'Vermont' as state, 2000 as year, 10 as val union all select 'Florida' as state, 2000, 5 as val
union all select 'Vermont', 2001, 16 union all select 'Florida' as state, 2001, 7
'''
;

def avg_year_percent_of_all_avg_year(field, dim)->round((avg(sum(field) by year, dim)/ avg(sum(field) by year )),2)::float::percent;

select @avg_year_percent_of_all_avg_year(val, state) as p_of_whole, state
order by p_of_whole desc;
"""
    )

    results = results.fetchall()
    assert int(results[0].p_of_whole * 100) == 68



from typing import Callable

from trilogy.core.constants import UNNEST_NAME
from trilogy.core.enums import Modifier, UnnestMode
from trilogy.core.models.build import (
    BuildComparison,
    BuildConcept,
    BuildConditional,
    BuildFunction,
    BuildParamaterizedConceptReference,
    BuildParenthetical,
)
from trilogy.core.models.execute import (
    CTE,
    InstantiatedUnnestJoin,
    Join,
    UnionCTE,
)


def render_unnest(
    unnest_mode: UnnestMode,
    quote_character: str,
    concept: BuildConcept | BuildParamaterizedConceptReference | BuildFunction,
    render_func: Callable[
        [BuildConcept | BuildParamaterizedConceptReference | BuildFunction, CTE], str
    ],
    cte: CTE,
):
    if not isinstance(concept, (BuildConcept, BuildParamaterizedConceptReference)):
        address = UNNEST_NAME
    else:
        address = concept.safe_address
    if unnest_mode == UnnestMode.CROSS_JOIN:
        return f"{render_func(concept, cte)} as {quote_character}{address}{quote_character}"
    elif unnest_mode == UnnestMode.CROSS_JOIN_UNNEST:
        return f"unnest({render_func(concept, cte)}) as {quote_character}{address}{quote_character}"
    elif unnest_mode == UnnestMode.PRESTO:
        return f"unnest({render_func(concept, cte)}) as t({quote_character}{UNNEST_NAME}{quote_character})"
    elif unnest_mode == UnnestMode.CROSS_JOIN_ALIAS:
        return f"{render_func(concept, cte)} as unnest_wrapper ({quote_character}{address}{quote_character})"
    elif unnest_mode == UnnestMode.SNOWFLAKE:
        # if we don't actually have a join, we're directly unnesting a concept, and we can skip the flatten
        if not cte.render_from_clause:
            return f"{render_func(concept, cte)} as unnest_wrapper ( unnest1, unnest2, unnest3, unnest4, {quote_character}{cte.join_derived_concepts[0].safe_address}{quote_character})"
        # otherwise, flatten the concept for the join
        return f"flatten({render_func(concept, cte)}) as unnest_wrapper ( unnest1, unnest2, unnest3, unnest4, {quote_character}{cte.join_derived_concepts[0].safe_address}{quote_character})"
    return f"{render_func(concept, cte)} as {quote_character}{address}{quote_character}"


def render_join_concept(
    name: str,
    quote_character: str,
    cte: CTE | UnionCTE,
    concept: BuildConcept,
    render_expr,
    inlined_ctes: set[str],
    use_map: dict[str, set[str]],
):
    if cte.name in inlined_ctes:
        base = render_expr(concept, cte)
        return base
    use_map[name].add(concept.address)
    return f"{quote_character}{name}{quote_character}.{quote_character}{concept.safe_address}{quote_character}"


def render_join(
    join: Join | InstantiatedUnnestJoin,
    quote_character: str,
    render_expr_func: Callable[
        [
            BuildConcept
            | BuildParamaterizedConceptReference
            | BuildFunction
            | BuildConditional
            | BuildComparison
            | BuildParenthetical,
            CTE,
        ],
        str,
    ],
    cte: CTE,
    use_map: dict[str, set[str]],
    null_wrapper: Callable[[str, str, list[Modifier]], str],
    unnest_mode: UnnestMode = UnnestMode.CROSS_APPLY,
) -> str | None:
    # {% for key in join.joinkeys %}{{ key.inner }} = {{ key.outer}}{% endfor %}
    if isinstance(join, InstantiatedUnnestJoin):
        if unnest_mode == UnnestMode.DIRECT:
            return None
        if not cte:
            raise ValueError("must provide a cte to build an unnest joins")
        if unnest_mode in (
            UnnestMode.CROSS_JOIN,
            UnnestMode.CROSS_JOIN_UNNEST,
            UnnestMode.CROSS_JOIN_ALIAS,
            UnnestMode.PRESTO,
        ):
            return f"CROSS JOIN {render_unnest(unnest_mode, quote_character, join.object_to_unnest, render_expr_func, cte)}"
        if unnest_mode == UnnestMode.SNOWFLAKE:
            return f"LEFT JOIN LATERAL {render_unnest(unnest_mode, quote_character, join.object_to_unnest, render_expr_func, cte)}"
        return f"FULL JOIN {render_unnest(unnest_mode, quote_character, join.object_to_unnest, render_expr_func, cte)}"
    # left_name = join.left_name
    right_name = join.right_name
    join.quote = quote_character
    # if cte.quote_address.get(join.right_name, False):
    #     join.quote = quote_character
    right_base = join.right_ref
    base_joinkeys = []
    if join.joinkey_pairs:
        base_joinkeys.extend(
            [
                null_wrapper(
                    render_join_concept(
                        join.get_name(pair.cte),
                        quote_character,
                        pair.cte,
                        pair.left,
                        render_expr_func,
                        join.inlined_ctes,
                        use_map=use_map,
                    ),
                    render_join_concept(
                        right_name,
                        quote_character,
                        join.right_cte,
                        pair.right,
                        render_expr_func,
                        join.inlined_ctes,
                        use_map=use_map,
                    ),
                    pair.modifiers
                    + (pair.left.modifiers or [])
                    + (pair.right.modifiers or [])
                    + (join.modifiers or []),
                )
                for pair in join.joinkey_pairs
            ]
        )
    if not base_joinkeys:
        base_joinkeys = ["1=1"]

    joinkeys = " AND ".join(sorted(base_joinkeys))
    base = f"{join.jointype.value.upper()} JOIN {right_base} on {joinkeys}"
    if join.condition:
        base = f"{base} and {render_expr_func(join.condition, cte)}"
    return base



from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import date, datetime
from functools import singledispatchmethod
from typing import Any

from jinja2 import Template

from trilogy.constants import DEFAULT_NAMESPACE, VIRTUAL_CONCEPT_PREFIX, MagicConstants
from trilogy.core.enums import (
    ConceptSource,
    DatePart,
    FunctionType,
    Modifier,
    Purpose,
    ValidationScope,
)
from trilogy.core.models.author import (
    AggregateWrapper,
    AlignClause,
    AlignItem,
    CaseElse,
    CaseWhen,
    Comment,
    Comparison,
    Concept,
    ConceptRef,
    Conditional,
    FilterItem,
    Function,
    FunctionCallWrapper,
    Grain,
    OrderBy,
    Ordering,
    OrderItem,
    Parenthetical,
    SubselectComparison,
    WhereClause,
    WindowItem,
)
from trilogy.core.models.core import (
    ArrayType,
    DataType,
    ListWrapper,
    MapWrapper,
    NumericType,
    TraitDataType,
    TupleWrapper,
)
from trilogy.core.models.datasource import (
    Address,
    ColumnAssignment,
    Datasource,
    Query,
    RawColumnExpr,
)
from trilogy.core.models.environment import Environment, Import
from trilogy.core.statements.author import (
    ArgBinding,
    ConceptDeclarationStatement,
    ConceptDerivationStatement,
    ConceptTransform,
    CopyStatement,
    FunctionDeclaration,
    ImportStatement,
    KeyMergeStatement,
    MergeStatementV2,
    MultiSelectStatement,
    PersistStatement,
    RawSQLStatement,
    RowsetDerivationStatement,
    SelectItem,
    SelectStatement,
    TypeDeclaration,
    ValidateStatement,
)

QUERY_TEMPLATE = Template(
    """{% if where %}WHERE
{{ where }}
{% endif %}SELECT{%- for select in select_columns %}
{{ select }},{% endfor %}{% if having %}
HAVING
{{ having }}
{% endif %}{%- if order_by %}
ORDER BY{% for order in order_by %}
{{ order }}{% if not loop.last %},{% endif %}{% endfor %}{% endif %}{%- if limit is not none %}
LIMIT {{ limit }}{% endif %}
;"""
)


@dataclass
class IndentationContext:
    """Tracks indentation state during rendering"""

    depth: int = 0
    indent_string: str = "    "  # 4 spaces by default

    @property
    def current_indent(self) -> str:
        return self.indent_string * self.depth

    def increase_depth(self, extra_levels: int = 1) -> "IndentationContext":
        return IndentationContext(
            depth=self.depth + extra_levels, indent_string=self.indent_string
        )


class Renderer:

    def __init__(
        self, environment: Environment | None = None, indent_string: str = "    "
    ):
        self.environment = environment
        self.indent_context = IndentationContext(indent_string=indent_string)

    @contextmanager
    def indented(self, levels: int = 1):
        """Context manager for temporarily increasing indentation"""
        old_context = self.indent_context
        self.indent_context = self.indent_context.increase_depth(levels)
        try:
            yield
        finally:
            self.indent_context = old_context

    def indent_lines(self, text: str, extra_levels: int = 0) -> str:
        """Apply current indentation to all lines in text"""
        if not text:
            return text

        indent = self.indent_context.indent_string * (
            self.indent_context.depth + extra_levels
        )
        lines = text.split("\n")
        indented_lines = []

        for line in lines:
            if line.strip():  # Only indent non-empty lines
                indented_lines.append(indent + line)
            else:
                indented_lines.append(line)  # Keep empty lines as-is

        return "\n".join(indented_lines)

    def render_statement_string(self, list_of_statements: list[Any]) -> str:
        new = []
        last_statement_type = None
        for stmt in list_of_statements:
            stmt_type = type(stmt)
            if last_statement_type is None:
                pass
            elif last_statement_type == Comment:
                new.append("\n")
            elif stmt_type != last_statement_type:
                new.append("\n\n")
            else:
                new.append("\n")
            new.append(self.to_string(stmt))
            last_statement_type = stmt_type
        return "".join(new)

    @singledispatchmethod
    def to_string(self, arg):
        raise NotImplementedError("Cannot render type {}".format(type(arg)))

    @to_string.register
    def _(self, arg: Environment):
        output_concepts = []
        constants: list[Concept] = []
        keys: list[Concept] = []
        properties = defaultdict(list)
        metrics = []
        # first, keys
        for concept in arg.concepts.values():
            if "__preql_internal" in concept.address:
                continue

            # don't render anything that came from an import
            if concept.namespace in arg.imports:
                continue
            if (
                concept.metadata
                and concept.metadata.concept_source == ConceptSource.AUTO_DERIVED
            ):
                continue
            elif not concept.lineage and concept.purpose == Purpose.CONSTANT:
                constants.append(concept)
            elif not concept.lineage and concept.purpose == Purpose.KEY:
                keys.append(concept)

            elif not concept.lineage and concept.purpose == Purpose.PROPERTY:
                if concept.keys:
                    # avoid duplicate declarations
                    # but we need better composite key support
                    for key in sorted(list(concept.keys))[:1]:
                        properties[key].append(concept)
                else:
                    keys.append(concept)
            else:
                metrics.append(concept)

        output_concepts = constants
        for key_concept in keys:
            output_concepts += [key_concept]
            output_concepts += properties.get(key_concept.name, [])
        output_concepts += metrics

        rendered_concepts = [
            self.to_string(ConceptDeclarationStatement(concept=concept))
            for concept in output_concepts
        ]

        rendered_datasources = [
            # extra padding between datasources
            # todo: make this more generic
            self.to_string(datasource) + "\n"
            for datasource in arg.datasources.values()
            if datasource.namespace == DEFAULT_NAMESPACE
        ]
        rendered_imports = []
        for _, imports in arg.imports.items():
            for import_statement in imports:
                rendered_imports.append(self.to_string(import_statement))
        components = []
        if rendered_imports:
            components.append(rendered_imports)
        if rendered_concepts:
            components.append(rendered_concepts)
        if rendered_datasources:
            components.append(rendered_datasources)

        final = "\n\n".join("\n".join(x) for x in components)
        return final

    @to_string.register
    def _(self, arg: TypeDeclaration):
        return f"type {arg.type.name} {self.to_string(arg.type.type)};"

    @to_string.register
    def _(self, arg: ArgBinding):
        if arg.default:
            return f"{arg.name}={self.to_string(arg.default)}"
        return f"{arg.name}"

    @to_string.register
    def _(self, arg: FunctionDeclaration):
        args = ", ".join([self.to_string(x) for x in arg.args])
        return f"def {arg.name}({args}) -> {self.to_string(arg.expr)};"

    @to_string.register
    def _(self, arg: Datasource):
        with self.indented():
            assignments = ",\n".join(
                [self.indent_lines(self.to_string(x)) for x in arg.columns]
            )

        if arg.non_partial_for:
            non_partial = f"\ncomplete where {self.to_string(arg.non_partial_for)}"
        else:
            non_partial = ""

        base = f"""datasource {arg.name} (
{assignments}
)
{self.to_string(arg.grain) if arg.grain.components else ''}{non_partial}
{self.to_string(arg.address)}"""

        if arg.where:
            base += f"\nwhere {self.to_string(arg.where)}"

        base += ";"
        return base

    @to_string.register
    def _(self, arg: "Grain"):
        final = []
        for comp in arg.components:
            if comp.startswith(DEFAULT_NAMESPACE):
                final.append(comp.split(".", 1)[1])
            else:
                final.append(comp)
        final = sorted(final)
        components = ",".join(x for x in final)
        return f"grain ({components})"

    @to_string.register
    def _(self, arg: "Query"):
        return f"""query {arg.text}"""

    @to_string.register
    def _(self, arg: RowsetDerivationStatement):
        return f"""rowset {arg.name} <- {self.to_string(arg.select)}"""

    @to_string.register
    def _(self, arg: "CaseWhen"):
        return (
            f"""WHEN {self.to_string(arg.comparison)} THEN {self.to_string(arg.expr)}"""
        )

    @to_string.register
    def _(self, arg: "CaseElse"):
        return f"""ELSE {self.to_string(arg.expr)}"""

    @to_string.register
    def _(self, arg: "FunctionCallWrapper"):
        args = [self.to_string(c) for c in arg.args]
        arg_string = ", ".join(args)
        return f"""@{arg.name}({arg_string})"""

    @to_string.register
    def _(self, arg: "Parenthetical"):
        return f"""({self.to_string(arg.content)})"""

    @to_string.register
    def _(self, arg: DataType):
        return arg.value

    @to_string.register
    def _(self, arg: "NumericType"):
        return f"""Numeric({arg.precision},{arg.scale})"""

    @to_string.register
    def _(self, arg: TraitDataType):
        traits = "::".join([x for x in arg.traits])
        return f"{self.to_string(arg.data_type)}::{traits}"

    @to_string.register
    def _(self, arg: ListWrapper):
        return "[" + ", ".join([self.to_string(x) for x in arg]) + "]"

    @to_string.register
    def _(self, arg: TupleWrapper):
        return "(" + ", ".join([self.to_string(x) for x in arg]) + ")"

    @to_string.register
    def _(self, arg: MapWrapper):
        def process_key_value(key, value):
            return f"{self.to_string(key)}: {self.to_string(value)}"

        return (
            "{"
            + ", ".join([process_key_value(key, value) for key, value in arg.items()])
            + "}"
        )

    @to_string.register
    def _(self, arg: DatePart):
        return arg.value

    @to_string.register
    def _(self, arg: "Address"):
        if arg.is_query:
            if arg.location.startswith("("):
                return f"query '''{arg.location[1:-1]}'''"
            return f"query '''{arg.location}'''"
        return f"address {arg.location}"

    @to_string.register
    def _(self, arg: "RawSQLStatement"):
        return f"raw_sql('''{arg.text}''');"

    @to_string.register
    def _(self, arg: "MagicConstants"):
        if arg == MagicConstants.NULL:
            return "null"
        return arg.value

    @to_string.register
    def _(self, arg: "ColumnAssignment"):
        if arg.modifiers:
            modifiers = "".join(
                [self.to_string(modifier) for modifier in sorted(arg.modifiers)]
            )
        else:
            modifiers = ""
        if isinstance(arg.alias, str):
            return f"{arg.alias}: {modifiers}{self.to_string(arg.concept)}"
        return f"{self.to_string(arg.alias)}: {modifiers}{self.to_string(arg.concept)}"

    @to_string.register
    def _(self, arg: "RawColumnExpr"):
        return f"raw('''{arg.text}''')"

    @to_string.register
    def _(self, arg: "ConceptDeclarationStatement"):
        concept = arg.concept
        if concept.metadata and concept.metadata.description:
            base_description = concept.metadata.description
        else:
            base_description = None
        if concept.namespace and concept.namespace != DEFAULT_NAMESPACE:
            namespace = f"{concept.namespace}."
        else:
            namespace = ""
        if not concept.lineage:
            if concept.purpose == Purpose.PROPERTY and concept.keys:
                if len(concept.keys) == 1:
                    output = f"{concept.purpose.value} {self.to_string(ConceptRef(address=list(concept.keys)[0]))}.{namespace}{concept.name} {self.to_string(concept.datatype)};"
                else:
                    keys = ",".join(
                        sorted(
                            list(
                                self.to_string(ConceptRef(address=x))
                                for x in concept.keys
                            )
                        )
                    )
                    output = f"{concept.purpose.value} <{keys}>.{namespace}{concept.name} {self.to_string(concept.datatype)};"
            else:
                output = f"{concept.purpose.value} {namespace}{concept.name} {self.to_string(concept.datatype)};"
        else:
            output = f"{concept.purpose.value} {namespace}{concept.name} <- {self.to_string(concept.lineage)};"
        if base_description:
            lines = "\n#".join(base_description.split("\n"))
            output += f" #{lines}"
        return output

    @to_string.register
    def _(self, arg: ArrayType):
        return f"list<{self.to_string(arg.value_data_type)}>"

    @to_string.register
    def _(self, arg: DataType):
        return arg.value

    @to_string.register
    def _(self, arg: date):
        return f"'{arg.isoformat()}'::date"

    @to_string.register
    def _(self, arg: datetime):
        return f"'{arg.isoformat()}'::datetime"

    @to_string.register
    def _(self, arg: ConceptDerivationStatement):
        # this is identical rendering;
        return self.to_string(ConceptDeclarationStatement(concept=arg.concept))

    @to_string.register
    def _(self, arg: PersistStatement):
        return f"PERSIST {arg.identifier} INTO {arg.address.location} FROM {self.to_string(arg.select)}"

    @to_string.register
    def _(self, arg: SelectItem):
        prefixes = []
        if Modifier.HIDDEN in arg.modifiers:
            prefixes.append("--")
        if Modifier.PARTIAL in arg.modifiers:
            prefixes.append("~")
        final = "".join(prefixes)
        return f"{final}{self.to_string(arg.content)}"

    @to_string.register
    def _(self, arg: ValidateStatement):
        targets = ",".join(arg.targets) if arg.targets else "*"
        if arg.scope.value == ValidationScope.ALL:
            return "validate all;"
        return f"validate {arg.scope.value} {targets};"

    @to_string.register
    def _(self, arg: SelectStatement):
        with self.indented():
            select_columns = [
                self.indent_lines(self.to_string(c)) for c in arg.selection
            ]
            where_clause = None
            if arg.where_clause:
                where_clause = self.indent_lines(self.to_string(arg.where_clause))
            having_clause = None
            if arg.having_clause:
                having_clause = self.indent_lines(self.to_string(arg.having_clause))
            order_by = None
            if arg.order_by:
                order_by = [
                    self.indent_lines(self.to_string(c)) for c in arg.order_by.items
                ]

        return QUERY_TEMPLATE.render(
            select_columns=select_columns,
            where=where_clause,
            having=having_clause,
            order_by=order_by,
            limit=arg.limit,
        )

    @to_string.register
    def _(self, arg: MultiSelectStatement):
        # Each select gets its own indentation
        select_parts = []
        for select in arg.selects:
            select_parts.append(
                self.to_string(select)[:-2]
            )  # Remove the trailing ";\n"

        base = "\nMERGE\n".join(select_parts)
        base += self.to_string(arg.align)
        if arg.where_clause:
            base += f"\nWHERE\n{self.to_string(arg.where_clause)}"
        if arg.order_by:
            base += f"\nORDER BY\n{self.to_string(arg.order_by)}"
        if arg.limit:
            base += f"\nLIMIT {arg.limit}"
        base += "\n;"
        return base

    @to_string.register
    def _(self, arg: CopyStatement):
        return f"COPY INTO {arg.target_type.value.upper()} '{arg.target}' FROM {self.to_string(arg.select)}"

    @to_string.register
    def _(self, arg: AlignClause):
        with self.indented():
            align_items = [self.indent_lines(self.to_string(c)) for c in arg.items]
        return "\nALIGN\n" + ",\n".join(align_items)

    @to_string.register
    def _(self, arg: AlignItem):
        return f"{arg.alias}:{','.join([self.to_string(c) for c in arg.concepts])}"

    @to_string.register
    def _(self, arg: OrderBy):
        with self.indented():
            order_items = [self.indent_lines(self.to_string(c)) for c in arg.items]
        return ",\n".join(order_items)

    @to_string.register
    def _(self, arg: Ordering):
        return arg.value

    @to_string.register
    def _(self, arg: "WhereClause"):
        base = f"{self.to_string(arg.conditional)}"
        if base[0] == "(" and base[-1] == ")":
            return base[1:-1]
        return base

    @to_string.register
    def _(self, arg: "Conditional"):
        return f"{self.to_string(arg.left)} {arg.operator.value} {self.to_string(arg.right)}"

    @to_string.register
    def _(self, arg: "SubselectComparison"):
        return f"{self.to_string(arg.left)} {arg.operator.value} {self.to_string(arg.right)}"

    @to_string.register
    def _(self, arg: "Comparison"):
        return f"{self.to_string(arg.left)} {arg.operator.value} {self.to_string(arg.right)}"

    @to_string.register
    def _(self, arg: "Comment"):
        lines = "\n#".join(arg.text.split("\n"))
        return f"{lines}"

    @to_string.register
    def _(self, arg: "WindowItem"):
        over = ",".join(self.to_string(c) for c in arg.over)
        order = ",".join(self.to_string(c) for c in arg.order_by)
        if over and order:
            return (
                f"{arg.type.value} {self.to_string(arg.content)} by {order} over {over}"
            )
        elif over:
            return f"{arg.type.value} {self.to_string(arg.content)} over {over}"
        return f"{arg.type.value} {self.to_string(arg.content)} by {order}"

    @to_string.register
    def _(self, arg: "FilterItem"):
        return f"filter {self.to_string(arg.content)} where {self.to_string(arg.where)}"

    @to_string.register
    def _(self, arg: "ConceptRef"):
        if arg.address == "__preql_internal.all_rows":
            return "*"
        if arg.name.startswith(VIRTUAL_CONCEPT_PREFIX) and self.environment:
            return self.to_string(self.environment.concepts[arg.address])

        ns, base = arg.address.rsplit(".", 1)
        if ns == DEFAULT_NAMESPACE:
            return base
        return arg.address

    @to_string.register
    def _(self, arg: "ImportStatement"):
        path: str = str(arg.path).replace("\\", ".")
        path = path.replace("/", ".")
        if path.endswith(".preql"):
            path = path.rsplit(".", 1)[0]
        if path.startswith("."):
            path = path[1:]
        if arg.alias == DEFAULT_NAMESPACE or not arg.alias:
            return f"import {path};"
        return f"import {path} as {arg.alias};"

    @to_string.register
    def _(self, arg: "Import"):
        path: str = str(arg.path).replace("\\", ".")
        path = path.replace("/", ".")
        if path.endswith(".preql"):
            path = path.rsplit(".", 1)[0]
        if path.startswith("."):
            path = path[1:]
        if arg.alias == DEFAULT_NAMESPACE or not arg.alias:
            return f"import {path};"
        return f"import {path} as {arg.alias};"

    @to_string.register
    def _(self, arg: "Concept"):
        if arg.name.startswith(VIRTUAL_CONCEPT_PREFIX):
            return self.to_string(arg.lineage)
        if arg.namespace == DEFAULT_NAMESPACE:
            return arg.name
        return arg.address

    @to_string.register
    def _(self, arg: "ConceptTransform"):
        return f"{self.to_string(arg.function)} -> {arg.output.name}"

    @to_string.register
    def _(self, arg: "Function"):
        args = [self.to_string(c) for c in arg.arguments]

        if arg.operator == FunctionType.SUBTRACT:
            return " - ".join(args)
        if arg.operator == FunctionType.ADD:
            return " + ".join(args)
        if arg.operator == FunctionType.MULTIPLY:
            return " * ".join(args)
        if arg.operator == FunctionType.DIVIDE:
            return " / ".join(args)
        if arg.operator == FunctionType.MOD:
            return f"{args[0]} % {args[1]}"
        if arg.operator == FunctionType.PARENTHETICAL:
            return f"({args[0]})"
        if arg.operator == FunctionType.GROUP:
            arg_string = ", ".join(args[1:])
            if len(args) == 1:
                return f"group({args[0]})"
            return f"group({args[0]}) by {arg_string}"

        if arg.operator == FunctionType.CONSTANT:
            return f"{', '.join(args)}"
        if arg.operator == FunctionType.CAST:
            return f"{self.to_string(arg.arguments[0])}::{self.to_string(arg.arguments[1])}"
        if arg.operator == FunctionType.INDEX_ACCESS:
            return f"{self.to_string(arg.arguments[0])}[{self.to_string(arg.arguments[1])}]"

        if arg.operator == FunctionType.CASE:
            with self.indented():
                indented_args = [
                    self.indent_lines(self.to_string(a)) for a in arg.arguments
                ]
            inputs = "\n".join(indented_args)
            return f"CASE\n{inputs}\n{self.indent_context.current_indent}END"

        if arg.operator == FunctionType.STRUCT:
            # zip arguments to pairs
            input_pairs = zip(arg.arguments[0::2], arg.arguments[1::2])
            with self.indented():
                pair_strings = []
                for k, v in input_pairs:
                    pair_line = f"{self.to_string(k)}-> {v}"
                    pair_strings.append(self.indent_lines(pair_line))
            inputs = ",\n".join(pair_strings)
            return f"struct(\n{inputs}\n{self.indent_context.current_indent})"
        if arg.operator == FunctionType.ALIAS:
            return f"{self.to_string(arg.arguments[0])}"
        inputs = ",".join(args)
        return f"{arg.operator.value}({inputs})"

    @to_string.register
    def _(self, arg: "OrderItem"):
        return f"{self.to_string(arg.expr)} {arg.order.value}"

    @to_string.register
    def _(self, arg: AggregateWrapper):
        if arg.by:
            by = ", ".join([self.to_string(x) for x in arg.by])
            return f"{self.to_string(arg.function)} by {by}"
        return f"{self.to_string(arg.function)}"

    @to_string.register
    def _(self, arg: MergeStatementV2):
        if len(arg.sources) == 1:
            return f"MERGE {self.to_string(arg.sources[0])} into {''.join([self.to_string(modifier) for modifier in arg.modifiers])}{self.to_string(arg.targets[arg.sources[0].address])};"
        return f"MERGE {arg.source_wildcard}.* into {''.join([self.to_string(modifier) for modifier in arg.modifiers])}{arg.target_wildcard}.*;"

    @to_string.register
    def _(self, arg: KeyMergeStatement):
        keys = ", ".join(sorted(list(arg.keys)))
        return f"MERGE PROPERTY <{keys}> from {arg.target.address};"

    @to_string.register
    def _(self, arg: Modifier):
        if arg == Modifier.PARTIAL:
            return "~"
        elif arg == Modifier.HIDDEN:
            return "--"
        elif arg == Modifier.NULLABLE:
            return "?"
        return arg.value

    @to_string.register
    def _(self, arg: int):
        return f"{arg}"

    @to_string.register
    def _(self, arg: str):
        return f"'{arg}'"

    @to_string.register
    def _(self, arg: float):
        return f"{arg}"

    @to_string.register
    def _(self, arg: bool):
        return f"{arg}"

    @to_string.register
    def _(self, arg: list):
        base = ", ".join([self.to_string(x) for x in arg])
        return f"[{base}]"


def render_query(query: "SelectStatement") -> str:
    return Renderer().to_string(query)


def render_environment(environment: "Environment") -> str:
    return Renderer().to_string(environment)



from dataclasses import dataclass
from typing import Literal, Union

from trilogy import Environment
from trilogy.ai.models import LLMMessage, LLMRequestOptions
from trilogy.ai.prompts import TRILOGY_LEAD_IN, create_query_prompt
from trilogy.ai.providers.base import LLMProvider
from trilogy.core.exceptions import (
    InvalidSyntaxException,
    NoDatasourceException,
    UndefinedConceptException,
    UnresolvableQueryException,
)
from trilogy.core.query_processor import process_query


@dataclass
class Conversation:

    messages: list[LLMMessage]
    provider: LLMProvider
    id: str | None = None

    @classmethod
    def create(
        cls,
        provider: LLMProvider,
        model_prompt: str = TRILOGY_LEAD_IN,
        id: str | None = None,
    ) -> "Conversation":
        system_message = LLMMessage(role="system", content=model_prompt)
        messages = [system_message]
        return cls(id=id, messages=messages, provider=provider)

    def add_message(
        self,
        message: Union[LLMMessage, str],
        role: Literal["user", "assistant"] = "user",
    ) -> None:
        """
        Add a message to the conversation.

        Args:
            message: Either an LLMMessage object or a string content
            role: The role for the message if a string is provided (default: 'user')
        """
        if isinstance(message, str):
            message = LLMMessage(role=role, content=message)
        self.messages.append(message)

    def get_response(self) -> LLMMessage:
        options = LLMRequestOptions()
        response = self.provider.generate_completion(options, history=self.messages)
        response_message = LLMMessage(role="assistant", content=response.text)
        self.add_message(response_message)
        return response_message

    def extract_response(self, content: str) -> str:
        # get contents in triple backticks
        content = content.replace('"""', "```")
        if "```" in content:
            parts = content.split("```")
            if len(parts) >= 3:
                return parts[-2].strip()
        return content

    def generate_query(
        self, user_input: str, environment: Environment, attempts: int = 4
    ) -> str:
        attempts = 0
        self.add_message(create_query_prompt(user_input, environment), role="user")
        e = None
        while attempts < 4:
            attempts += 1

            response_message = self.get_response()
            response = self.extract_response(response_message.content)
            if not response.strip():
                self.add_message(
                    "Your response did not contain a valid SQL query. Please provide a valid SQL query enclosed in triple backticks.",
                    role="user",
                )
                continue
            if not response.strip()[-1] == ";":
                response += ";"
            try:
                _, raw = environment.parse(response)
                process_query(statement=raw[-1], environment=environment)
                return response
            except (
                InvalidSyntaxException,
                NoDatasourceException,
                UnresolvableQueryException,
                UndefinedConceptException,
                SyntaxError,
            ) as e2:
                e = e2
                self.add_message(
                    f"Your extracted response - {response} - could not be parsed due to the error: {str(e)}. Please generate a new query with the issues fixed. Use the same response format.",
                    role="user",
                )
        raise Exception(
            f"Failed to generate a valid query after {attempts} attempts. Last error: {str(e)}. Full conversation: {self.messages}"
        )



from typing import List

from trilogy.constants import logger
from trilogy.core.models.build import (
    BuildConcept,
    BuildGrain,
    BuildWhereClause,
    BuildWindowItem,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.node_generators.common import (
    gen_enrichment_node,
)
from trilogy.core.processing.nodes import (
    History,
    StrategyNode,
    WhereSafetyNode,
    WindowNode,
)
from trilogy.core.processing.utility import create_log_lambda, padding
from trilogy.utility import unique

LOGGER_PREFIX = "[GEN_WINDOW_NODE]"


WINDOW_TYPES = (BuildWindowItem,)


def resolve_window_parent_concepts(
    concept: BuildConcept, environment: BuildEnvironment, depth: int
) -> tuple[BuildConcept, List[BuildConcept]]:
    if not isinstance(concept.lineage, WINDOW_TYPES):
        raise ValueError
    base = []
    if concept.lineage.over:
        base += concept.lineage.over
    if concept.lineage.order_by:
        for item in concept.lineage.order_by:
            base += item.concept_arguments
    if concept.grain:
        for gitem in concept.grain.components:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} appending grain item {gitem} to base"
            )
            base.append(environment.concepts[gitem])
    return concept.lineage.content, unique(base, "address")


def gen_window_node(
    concept: BuildConcept,
    local_optional: list[BuildConcept],
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    history: History,
    conditions: BuildWhereClause | None = None,
) -> StrategyNode | None:
    base, parent_concepts = resolve_window_parent_concepts(concept, environment, depth)
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} generating window node for {concept} with parents {[x.address for x in parent_concepts]} and optional {local_optional}"
    )
    equivalent_optional = [
        x
        for x in local_optional
        if isinstance(x.lineage, WINDOW_TYPES)
        and resolve_window_parent_concepts(x, environment, depth)[1] == parent_concepts
    ]

    targets = [base]
    # append in keys to get the right grain
    if concept.keys:
        for item in concept.keys:
            if item in targets:
                continue
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} appending search for key {item}"
            )
            targets.append(environment.concepts[item])
    additional_outputs = []
    if equivalent_optional:
        for x in equivalent_optional:
            assert isinstance(x.lineage, WINDOW_TYPES)
            base, parents = resolve_window_parent_concepts(x, environment, depth)
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} found equivalent optional {x} with parents {parents}"
            )
            additional_outputs.append(x)
            # also append the base concept it's being grouped over
            targets.append(base)

    grain_equivalents = [
        x
        for x in local_optional
        if x.keys
        and all([key in targets for key in x.keys])
        and x.grain == concept.grain
    ]

    for x in grain_equivalents:
        if x.address in additional_outputs:
            continue
        targets.append(x)

    # finally, the ones we'll need to enrich
    non_equivalent_optional = [x for x in local_optional if x.address not in targets]

    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} resolving final parents {parent_concepts + targets}"
    )
    parent_node: StrategyNode = source_concepts(
        mandatory_list=parent_concepts + targets,
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
        conditions=conditions,
    )
    if not parent_node:
        logger.info(f"{padding(depth)}{LOGGER_PREFIX} window node parents unresolvable")
        return None
    parent_node.resolve()
    if not all(
        [
            x.address in [y.address for y in parent_node.output_concepts]
            for x in parent_concepts
        ]
    ):
        missing = [
            x
            for x in parent_concepts
            if x.address not in [y.address for y in parent_node.output_concepts]
        ]
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} window node parents unresolvable, missing {missing}"
        )
        raise SyntaxError
    _window_node = WindowNode(
        input_concepts=parent_concepts + targets,
        output_concepts=[concept] + additional_outputs + parent_concepts + targets,
        environment=environment,
        parents=[
            parent_node,
        ],
        depth=depth,
        preexisting_conditions=conditions.conditional if conditions else None,
    )
    _window_node.rebuild_cache()
    _window_node.resolve()

    window_node = WhereSafetyNode(
        input_concepts=[concept] + additional_outputs + parent_concepts + targets,
        output_concepts=[concept] + additional_outputs + parent_concepts + targets,
        environment=environment,
        parents=[_window_node],
        preexisting_conditions=conditions.conditional if conditions else None,
        grain=BuildGrain.from_concepts(
            concepts=[concept] + additional_outputs + parent_concepts + targets,
            environment=environment,
        ),
    )
    if not non_equivalent_optional:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no optional concepts, returning window node"
        )
        # prune outputs if we don't need join keys
        window_node.set_output_concepts([concept] + additional_outputs + targets)
        return window_node

    missing_optional = [
        x.address
        for x in local_optional
        if x.address not in window_node.output_concepts
    ]

    if not missing_optional:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no extra enrichment needed for window node, has all of {[x.address for x in local_optional]}"
        )
        return window_node
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} window node for {concept.address} requires enrichment, missing {missing_optional}, has {[x.address for x in window_node.output_concepts]}"
    )

    return gen_enrichment_node(
        window_node,
        join_keys=[
            environment.concepts[c]
            for c in BuildGrain.from_concepts(
                concepts=targets, environment=environment
            ).components
        ],
        local_optional=local_optional,
        environment=environment,
        g=g,
        depth=depth,
        source_concepts=source_concepts,
        log_lambda=create_log_lambda(LOGGER_PREFIX, depth, logger),
        history=history,
        conditions=conditions,
    )



# from trilogy.compiler import compile
from datetime import date, datetime
from logging import INFO

from pytest import raises

from trilogy import Dialects
from trilogy.constants import logger
from trilogy.core.enums import Derivation, Purpose
from trilogy.core.exceptions import InvalidSyntaxException
from trilogy.core.models.core import ArrayType, DataType
from trilogy.core.models.environment import Environment
from trilogy.core.query_processor import process_query
from trilogy.core.statements.author import SelectStatement
from trilogy.dialect.base import BaseDialect
from trilogy.dialect.bigquery import BigqueryDialect
from trilogy.dialect.duckdb import DuckDBDialect
from trilogy.dialect.snowflake import SnowflakeDialect
from trilogy.dialect.sql_server import SqlServerDialect
from trilogy.parser import parse

logger.setLevel(INFO)

TEST_DIALECTS = [
    BaseDialect(),
    BigqueryDialect(),
    DuckDBDialect(),
    SqlServerDialect(),
    SnowflakeDialect(),
]


def test_typing():
    x = ArrayType(type=DataType.INTEGER)

    assert x in set([x])


def test_negative_float_parsing(test_environment):
    declarations = """

    const x <- -1.23;

select
x
where x >-1.24;

    """
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]

    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select, hooks=[]))


def test_functions(test_environment):
    declarations = """

select
    total_revenue,
    min_order_id,
    max_order_id,
    order_count,
    distinct_order_count
;


    """
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]

    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select, hooks=[]))


def test_wrapped_property_functions(test_environment):
    declarations = """

select
    product_id,
    avg(category_name_length) ->average_category_name_length
;


    """
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]

    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select))


def test_window_functions(test_environment):
    declarations = """

    property prior_order_id <- lag order_id by order_id desc;
    property next_order_id <- lead order_id order by order_id asc;
    select
        order_id,
        prior_order_id,
        next_order_id
    ;
    


        """
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]

    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select))


def test_window_datatype(test_environment):
    declarations = """

    auto category_rank <- rank category_name order by count(order_id) desc;

        """
    env, parsed = parse(declarations, environment=test_environment)
    assert env.concepts["category_rank"].datatype == DataType.INTEGER


def test_date_functions(test_environment):
    declarations = """

    select
        order_id,
        order_timestamp,
        date(order_timestamp) -> order_date,
        datetime(order_timestamp) -> order_timestamp_datetime,
        timestamp(order_timestamp) -> order_timestamp_dos,
        second(order_timestamp) -> order_second,
        minute(order_timestamp) -> order_minute,
        hour(order_timestamp) -> order_hour,
        day(order_timestamp) -> order_day,
        week(order_timestamp) -> order_week,
        month(order_timestamp) -> order_month,
        quarter(order_timestamp) -> order_quarter,
        year(order_timestamp) -> order_year,
        date_trunc(order_timestamp, month) -> order_month_trunc,
        date_add(order_timestamp, month, 1) -> one_month_post_order,
        date_sub(order_timestamp, month, 1) -> one_month_pre_order,
        date_trunc(order_timestamp, day) -> order_day_trunc,
        date_trunc(order_timestamp, year) -> order_year_trunc,
        date_trunc(order_timestamp, hour) -> order_hour_trunc,
        date_trunc(order_timestamp, minute) -> order_minute_trunc,
        date_trunc(order_timestamp, second) -> order_second_trunc,
        date_trunc(order_timestamp, quarter) -> order_quarter_trunc,
        date_trunc(order_timestamp, week) -> order_week_trunc,
        date_part(order_timestamp, month) -> order_month_part,
        date_part(order_timestamp, day) -> order_day_part,
        date_part(order_timestamp, year) -> order_year_part,
        date_part(order_timestamp, hour) -> order_hour_part,
        date_part(order_timestamp, minute) -> order_minute_part,
        date_part(order_timestamp, second) -> order_second_part,
        date_part(order_timestamp, quarter) -> order_quarter_part,
        date_part(order_timestamp, week) -> order_week_part,
        date_part(order_timestamp, day_of_week) -> order_day_of_week_part,
        month_name(order_timestamp) -> order_month_name,
        day_name(order_timestamp) -> order_day_name
    ;
    
    
        """
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]

    for dialect in TEST_DIALECTS:
        engine = None
        if isinstance(dialect, DuckDBDialect):
            engine = Dialects.DUCK_DB.default_executor(environment=test_environment)
        dialect.compile_statement(process_query(test_environment, select))
        if engine:
            engine.execute_raw_sql(
                """CREATE TABLE tblRevenue AS
SELECT 
    100.50 as revenue,
    1 as order_id,
    101 as product_id,
    TIMESTAMP '2024-01-15 10:30:00' as order_timestamp
UNION ALL
SELECT 
    250.75 as revenue,
    2 as order_id,
    102 as product_id,
    TIMESTAMP '2024-01-16 14:45:30' as order_timestamp;"""
            )
            results = engine.execute_query(select)
            assert results
            assert results.fetchall()


def test_bad_cast(test_environment):
    declarations = """

    select
        order_id,
        date(order_id) -> order_id_date,
    ;"""
    with raises(InvalidSyntaxException):
        env, parsed = parse(declarations, environment=test_environment)


def test_explicit_cast(test_environment):
    declarations = """
    property _str_order_id <- cast(order_id as string);
    select
        order_id,
        _str_order_id
    ;"""
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]
    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select))


def test_literal_cast(test_environment):
    declarations = """
select
    '1'::int -> one,
    '1'::float -> one_float,
    '1'::string -> one_string,
    '2024-01-01'::date -> one_date,
    '2024-01-01 01:01:01'::datetime -> one_datetime,
    'true'::bool -> one_bool,
    ;"""
    env, parsed = parse(declarations, environment=test_environment)

    select: SelectStatement = parsed[-1]
    z = (
        Dialects.DUCK_DB.default_executor(environment=test_environment)
        .execute_query(parsed[-1])
        .fetchall()
    )
    assert z[0].one == 1
    assert z[0].one_float == 1.0
    assert z[0].one_string == "1"
    assert z[0].one_date == date(year=2024, month=1, day=1)
    assert z[0].one_datetime == datetime(
        year=2024, month=1, day=1, hour=1, minute=1, second=1
    )
    assert z[0].one_bool
    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select))


def test_math_functions(test_environment):
    declarations = """
    
    
    property inflated_order_value<- multiply(revenue, 2);
    property fixed_order_value<- inflated_order_value / 2;
    property order_add <- revenue + 2;
    property order_sub <- revenue - 2;
    property order_id.order_nested <- revenue * 2/2;
    property order_id.rounded <- round(revenue + 2.01,2);
    property order_id.rounded_default <- round(revenue + 2.01);
    property order_id.floor <- floor(revenue + 2.01);
    property order_id.ceil <- ceil(revenue + 2.01);
    constant random <- random(1);
    select
        order_id,
        inflated_order_value,
        order_nested,
        fixed_order_value,
        order_sub,
        order_add,
        rounded,
        rounded_default,
        floor,
        ceil,
        random,
    ;"""
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]
    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select))


def test_random_randomness(test_environment):
    declarations = """
    auto x <- unnest([1,2,3,4]);
    select
        x,
        random(x) -> random_1,
    ;"""
    env, parsed = parse(declarations, environment=test_environment)
    z = (
        Dialects.DUCK_DB.default_executor(environment=test_environment)
        .execute_query(parsed[-1])
        .fetchall()
    )
    assert z[0].random_1 != z[1].random_1, f"{z[0].random_1} == {z[1].random_1}"


def test_string_functions(test_environment):
    declarations = """
    property test_name <- concat(category_name, '_test');
    property upper_name <- upper(category_name);
    property lower_name <- lower(category_name);
    property substring_name <- substring(category_name, 1, 3);
    property strpos_name <- strpos(category_name, 'a');
    property like_name <- like(category_name, 'a%');
    property like_alt <- category_name like 'a%';
    property regex_contains <- regexp_contains(category_name, 'a');
    property regex_substring <- regexp_extract(category_name, 'a');
    property regex_replace <- regexp_replace(category_name, 'a', 'b');

    select
        category_id,
        test_name,
        upper_name,
        lower_name,
        substring_name,
        strpos_name,
        like_name,
        like_alt,
        regex_contains,
        regex_substring,
        regex_replace,
        hash(category_name, md5) -> hash_md5,
        hash(category_name, sha1) -> hash_sha1,
        hash(category_name, sha256) -> hash_sha256,
        # hash(category_name, sha512) -> hash_sha512
    ;"""
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]
    for dialect in TEST_DIALECTS:
        engine = None
        if isinstance(dialect, DuckDBDialect):
            engine = Dialects.DUCK_DB.default_executor(environment=test_environment)
        dialect.compile_statement(process_query(test_environment, select))
        if engine:
            engine.execute_raw_sql(
                """CREATE TABLE tblCategory AS
SELECT 
    2 category_id,
    'category_a' as category_name
UNION ALL
SELECT 
    3 category_id,
    'category_b' as category_name;"""
            )
            results = engine.execute_query(select)
            assert results
            assert results.fetchall()


def test_case_function(test_environment):
    declarations = """
    property test_upper_case <- CASE WHEN category_name = upper(category_name) then True else False END;
    select
        category_name,
        test_upper_case
    ;"""
    env, parsed = parse(declarations, environment=test_environment)

    assert (
        test_environment.concepts["category_name"]
        in test_environment.concepts["test_upper_case"].lineage.concept_arguments
    )
    select: SelectStatement = parsed[-1]
    for dialect in TEST_DIALECTS:
        compiled = dialect.compile_statement(process_query(test_environment, select))
        assert "CASE" in compiled
        assert "ELSE" in compiled
        assert "END" in compiled
        assert test_environment.concepts["test_upper_case"].datatype == DataType.BOOL


def test_case_like_function(test_environment):
    declarations = """
    property test_like <- CASE WHEN category_name like '%abc%' then True else False END;
    select
        category_name,
        test_like
    ;"""
    env, parsed = parse(declarations, environment=test_environment)
    assert (
        test_environment.concepts["category_name"]
        in test_environment.concepts["test_like"].lineage.concept_arguments
    )
    select: SelectStatement = parsed[-1]
    for dialect in TEST_DIALECTS:
        compiled = dialect.compile_statement(process_query(test_environment, select))
        assert "CASE" in compiled
        assert "ELSE" in compiled
        assert "END" in compiled
        assert test_environment.concepts["test_like"].datatype == DataType.BOOL


def test_split_and_index_function(test_environment):
    declarations = """
    constant test_string <- 'abc_def';

    
    select
        split(test_string, '_')->split_string,
        split(test_string, '_')[0] -> first_element,
    ;"""
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]
    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select))


def test_coalesce(test_environment):
    declarations = """
    constant test_string <- 'abc_def';

    
    select
        coalesce(test_string, 'test')->coalesce_string,
        coalesce(null, 'test')->coalesce_null,
        coalesce(null, null, test_string)->coalesce_null_null,
    ;"""
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]
    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select))


def test_constants(test_environment):
    declarations = """
    const current_date <- current_date();
    const current_datetime <- current_datetime();
    
    select
        current_date,
        current_datetime,
    ;"""
    env, parsed = parse(declarations, environment=test_environment)
    select: SelectStatement = parsed[-1]
    for dialect in TEST_DIALECTS:
        dialect.compile_statement(process_query(test_environment, select))


def test_unnest(test_environment):
    declarations = """
   auto int_list <- [1,2,3,4];
    auto x <- unnest(int_list);
    
    select
        x
    ;"""
    env, parsed = parse(declarations, environment=test_environment)
    assert env.concepts["int_list"].datatype == ArrayType(type=DataType.INTEGER)
    assert env.concepts["x"].datatype == DataType.INTEGER


def test_validate_constant_functions():
    x = Environment()
    env, _ = x.parse(
        """
            const current_date <- current_date();
            """
    )
    assert env.concepts["current_date"].purpose == Purpose.CONSTANT
    assert env.concepts["current_date"].derivation == Derivation.CONSTANT



from trilogy.constants import Rendering
from trilogy.dialect.base import BaseDialect
from trilogy.dialect.enums import Dialects


def get_dialect_generator(
    dialect: Dialects, rendering: Rendering | None = None
) -> BaseDialect:
    if dialect == Dialects.BIGQUERY:
        from trilogy.dialect.bigquery import BigqueryDialect

        return BigqueryDialect(rendering=rendering)
    elif dialect == Dialects.SQL_SERVER:
        from trilogy.dialect.sql_server import SqlServerDialect

        return SqlServerDialect(rendering=rendering)
    elif dialect == Dialects.DUCK_DB:
        from trilogy.dialect.duckdb import DuckDBDialect

        return DuckDBDialect(rendering=rendering)
    elif dialect == Dialects.PRESTO:
        from trilogy.dialect.presto import PrestoDialect

        return PrestoDialect(rendering=rendering)
    elif dialect == Dialects.TRINO:
        from trilogy.dialect.presto import TrinoDialect

        return TrinoDialect(rendering=rendering)
    elif dialect == Dialects.POSTGRES:
        from trilogy.dialect.postgres import PostgresDialect

        return PostgresDialect(rendering=rendering)
    elif dialect == Dialects.SNOWFLAKE:
        from trilogy.dialect.snowflake import SnowflakeDialect

        return SnowflakeDialect(rendering=rendering)
    elif dialect == Dialects.DATAFRAME:
        from trilogy.dialect.dataframe import DataframeDialect

        return DataframeDialect(rendering=rendering)
    else:
        raise ValueError(f"Unsupported dialect {dialect}")



from pathlib import Path

from trilogy import Dialects
from trilogy.core.enums import FunctionType
from trilogy.core.statements.execute import ProcessedCopyStatement
from trilogy.parser import parse

# from trilogy.compiler import compile


def test_declarations():
    declarations = """key user_id int metadata(description="the description");
property user_id.display_name string metadata(description="The display name ");
property user_id.about_me string metadata(description="User provided description");
key post_id int;
    """
    parse(declarations)


def test_duplicate_declarations():
    declarations = """key user_id int metadata(description="the description");
property user_id.display_name string metadata(description="The display name ");
property user_id.about_me string metadata(description="User provided description");
key post_id int;
key post_id int;
    """
    try:
        parse(declarations)
    except Exception as e:
        assert "line 5" in str(e)


def test_datasource():
    text = """key user_id int metadata(description="the description");
property user_id.display_name string metadata(description="The display name ");
property user_id.about_me string metadata(description="User provided description");
key post_id int;


datasource posts (
    user_id: user_id,
    id: post_id
    )
    grain (post_id)
    address `bigquery-public-data.stackoverflow.post_history`
;
"""
    parse(text)


def test_io_statement():
    from trilogy.hooks.query_debugger import DebuggingHook

    DebuggingHook()
    target = Path(__file__).parent / "test_io_statement.csv"
    if target.exists():
        target.unlink()
    text = f"""const array <- [1,2,3,4];

auto x <- unnest(array);

copy into csv '{target}' from select x -> test order by test asc;
"""
    exec = Dialects.DUCK_DB.default_executor()
    results = exec.parse_text(text)
    assert exec.environment.concepts["x"].lineage.operator == FunctionType.UNNEST
    assert isinstance(results[-1], ProcessedCopyStatement)
    for z in results:
        exec.execute_query(z)
    assert target.exists(), "csv file was not created"


def test_datasource_where():
    text = """key user_id int metadata(description="the description");
property user_id.display_name string metadata(description="The display name ");
property user_id.about_me string metadata(description="User provided description");
key post_id int;


datasource x_posts (
    user_id: user_id,
    id: post_id
    )
    grain (post_id)
    address `bigquery-public-data.stackoverflow.post_history`
    where post_id = 2
;
"""
    parse(text)



from trilogy.ai.conversation import Conversation
from trilogy.ai.enums import Provider
from trilogy.ai.execute import text_to_query
from trilogy.ai.models import LLMMessage
from trilogy.ai.prompts import create_query_prompt
from trilogy.ai.providers.anthropic import AnthropicProvider
from trilogy.ai.providers.google import GoogleProvider
from trilogy.ai.providers.openai import OpenAIProvider

__all__ = [
    "Conversation",
    "LLMMessage",
    "OpenAIProvider",
    "GoogleProvider",
    "AnthropicProvider",
    "create_query_prompt",
    "text_to_query",
    "Provider",
]



from .base_optimization import OptimizationRule
from .hide_unused_concept import HideUnusedConcepts
from .inline_datasource import InlineDatasource
from .predicate_pushdown import PredicatePushdown, PredicatePushdownRemove

__all__ = [
    "OptimizationRule",
    "InlineDatasource",
    "PredicatePushdown",
    "PredicatePushdownRemove",
    "HideUnusedConcepts",
]



from trilogy import parse


def test_derivations(test_environment):
    assert (
        test_environment.concepts["order_timestamp"].address == "local.order_timestamp"
    )
    assert (
        test_environment.concepts["order_timestamp.date"].address
        == "local.order_timestamp.date"
    )


def test_hour_derivation(test_environment):
    hour_derived = test_environment.concepts["order_timestamp.hour"]
    assert "hour" in hour_derived.datatype.traits, hour_derived

    year_derived = test_environment.concepts["order_timestamp.year"]
    assert "year" in year_derived.datatype.traits, year_derived


def test_filtering_where_on_derived_aggregate(test_environment):
    exception = False
    try:
        env, _ = parse(
            """key x int;
    property x.cost float;

    datasource x_source (
        x:x,
        cost:cost)
        grain(x)
        address x_source;


    SELECT
        sum(cost)-> filtered_cst
    where
    x > 10 and filtered_cst >1000;
        """
        )
    except Exception as e:
        exception = True
        assert (
            "Cannot reference an aggregate derived in the select (local.filtered_cst)"
            in str(e)
        )
    assert exception, "should have an exception"


def test_filtering_having_on_unincluded_value(test_environment):
    exception = False
    try:
        env, _ = parse(
            """key x int;
    property x.cost float;

    datasource x_source (
        x:x,
        cost:cost)
        grain(x)
        address x_source;


    SELECT
        sum(cost)-> filtered_cst
    having
    x > 10 and filtered_cst >1000;
        """
        )
    except Exception as e:
        exception = True
        assert str(e).startswith(
            "Cannot reference a column (local.x) that is not in the select projection in the HAVING clause, move to WHERE"
        ), str(e)
    assert exception, "should have an exception"


def test_filtering_valid(test_environment):
    env, _ = parse(
        """key x int;
property x.cost float;

datasource x_source (
    x:x,
    cost:cost)
    grain(x)
    address x_source;

where
    x>-10
SELECT
    sum(cost)-> filtered_cst
having
    filtered_cst >1000;
    """
    )



from collections import defaultdict
from typing import Callable, List, Tuple

from trilogy.core.enums import Derivation, Purpose
from trilogy.core.models.build import (
    BuildAggregateWrapper,
    BuildComparison,
    BuildConcept,
    BuildFilterItem,
    BuildFunction,
    BuildWhereClause,
    LooseBuildConceptList,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.nodes import (
    History,
    NodeJoin,
)
from trilogy.core.processing.nodes.base_node import StrategyNode
from trilogy.core.processing.nodes.merge_node import MergeNode
from trilogy.utility import unique

AGGREGATE_TYPES = (BuildAggregateWrapper,)
FUNCTION_TYPES = (BuildFunction,)


def resolve_function_parent_concepts(
    concept: BuildConcept, environment: BuildEnvironment
) -> List[BuildConcept]:
    if not isinstance(
        concept.lineage, (*FUNCTION_TYPES, *AGGREGATE_TYPES, BuildComparison)
    ):
        raise ValueError(
            f"Concept {concept} lineage is not function or aggregate, is {type(concept.lineage)}"
        )
    if concept.derivation == Derivation.AGGREGATE:
        base: list[BuildConcept] = []
        if not concept.grain.abstract:
            base = concept.lineage.concept_arguments + [
                environment.concepts[c] for c in concept.grain.components
            ]
            # if the base concept being aggregated is a property with a key
            # keep the key as a parent
        else:
            base = concept.lineage.concept_arguments
        if isinstance(concept.lineage, AGGREGATE_TYPES):
            # for aggregate wrapper, don't include the by
            extra_property_grain = concept.lineage.function.concept_arguments
        else:
            extra_property_grain = concept.lineage.concept_arguments
        for x in extra_property_grain:
            if isinstance(x, BuildConcept) and x.purpose == Purpose.PROPERTY and x.keys:
                base += [environment.concepts[c] for c in x.keys]
        return unique(base, "address")
    # TODO: handle basic lineage chains?
    return unique(concept.lineage.concept_arguments, "address")


def resolve_condition_parent_concepts(
    condition: BuildWhereClause,
) -> Tuple[List[BuildConcept], List[Tuple[BuildConcept, ...]]]:
    base_existence = []
    base_rows: list[BuildConcept] = []
    base_rows += condition.row_arguments
    for ctuple in condition.existence_arguments:
        base_existence.append(ctuple)
    return unique(base_rows, "address"), base_existence


def resolve_filter_parent_concepts(
    concept: BuildConcept,
    environment: BuildEnvironment,
) -> Tuple[List[BuildConcept], List[Tuple[BuildConcept, ...]]]:
    if not isinstance(concept.lineage, (BuildFilterItem,)):
        raise ValueError(
            f"Concept {concept} lineage is not filter item, is {type(concept.lineage)}"
        )
    direct_parent = concept.lineage.content
    base_existence = []
    base_rows = [direct_parent] if isinstance(direct_parent, BuildConcept) else []
    condition_rows, condition_existence = resolve_condition_parent_concepts(
        concept.lineage.where
    )
    base_rows += condition_rows
    base_existence += condition_existence
    # this is required so that
    if (
        isinstance(direct_parent, BuildConcept)
        and direct_parent.purpose in (Purpose.PROPERTY, Purpose.METRIC)
        and direct_parent.keys
    ):
        base_rows += [environment.concepts[c] for c in direct_parent.keys]

    if concept.lineage.where.existence_arguments:
        return (
            unique(base_rows, "address"),
            base_existence,
        )
    return unique(base_rows, "address"), []


def gen_property_enrichment_node(
    base_node: StrategyNode,
    extra_properties: list[BuildConcept],
    history: History,
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    log_lambda: Callable,
    conditions: BuildWhereClause | None = None,
):
    required_keys: dict[str, set[str]] = defaultdict(set)
    for x in extra_properties:
        if not x.keys:
            raise SyntaxError(f"Property {x.address} missing keys in lookup")
        keys = "-".join([y for y in x.keys])
        required_keys[keys].add(x.address)
    final_nodes = []
    for _k, vs in required_keys.items():
        log_lambda(f"Generating enrichment node for {_k} with {vs}")
        ks = _k.split("-")
        enrich_node: StrategyNode = source_concepts(
            mandatory_list=[environment.concepts[k] for k in ks]
            + [environment.concepts[v] for v in vs],
            environment=environment,
            g=g,
            depth=depth + 1,
            history=history,
            conditions=conditions,
        )
        final_nodes.append(enrich_node)
    return MergeNode(
        input_concepts=unique(
            base_node.output_concepts
            + extra_properties
            + [
                environment.concepts[v]
                for k, values in required_keys.items()
                for v in values
            ],
            "address",
        ),
        output_concepts=base_node.output_concepts + extra_properties,
        environment=environment,
        parents=[
            base_node,
        ]
        + final_nodes,
        preexisting_conditions=conditions.conditional if conditions else None,
    )


def gen_enrichment_node(
    base_node: StrategyNode,
    join_keys: List[BuildConcept],
    local_optional: list[BuildConcept],
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    log_lambda,
    history: History,
    conditions: BuildWhereClause | None = None,
):
    local_opts = LooseBuildConceptList(concepts=local_optional)

    extra_required = [
        x
        for x in local_opts
        if x not in base_node.output_lcl or x in base_node.partial_lcl
    ]

    # property lookup optimization
    # this helps create ergonomic merge nodes when evaluating a normalized star schema
    # as we only want to lookup the missing properties based on the relevant keys
    if all([x.purpose == Purpose.PROPERTY for x in extra_required]):
        if all(
            x.keys and all([key in base_node.output_lcl for key in x.keys])
            for x in extra_required
        ):
            log_lambda(
                f"{str(type(base_node).__name__)} returning property optimized enrichment node for {extra_required[0].keys}"
            )
            return gen_property_enrichment_node(
                base_node,
                extra_required,
                environment=environment,
                g=g,
                depth=depth,
                source_concepts=source_concepts,
                history=history,
                conditions=conditions,
                log_lambda=log_lambda,
            )
    log_lambda(
        f"{str(type(base_node).__name__)} searching for join keys {LooseBuildConceptList(concepts=join_keys)} and extra required {local_opts}"
    )
    enrich_node: StrategyNode = source_concepts(  # this fetches the parent + join keys
        # to then connect to the rest of the query
        mandatory_list=join_keys + extra_required,
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
        conditions=conditions,
    )
    if not enrich_node:
        log_lambda(
            f"{str(type(base_node).__name__)} enrichment node unresolvable, returning just group node"
        )
        return base_node
    log_lambda(
        f"{str(type(base_node).__name__)} returning merge node with group node + enrichment node"
    )
    non_hidden = [
        x
        for x in base_node.output_concepts
        if x.address not in base_node.hidden_concepts
    ]
    return MergeNode(
        input_concepts=unique(join_keys + extra_required + non_hidden, "address"),
        output_concepts=unique(join_keys + extra_required + non_hidden, "address"),
        environment=environment,
        parents=[enrich_node, base_node],
        force_group=False,
        preexisting_conditions=conditions.conditional if conditions else None,
        depth=depth,
    )


def resolve_join_order(joins: List[NodeJoin]) -> List[NodeJoin]:
    if not joins:
        return []
    available_aliases: set[StrategyNode] = set()
    final_joins_pre = [*joins]
    final_joins = []
    left = set()
    right = set()
    for join in joins:
        left.add(join.left_node)
        right.add(join.right_node)

    potential_basis = left.difference(right)
    base_candidates = [x for x in final_joins_pre if x.left_node in potential_basis]
    if not base_candidates:
        raise SyntaxError(
            f"Unresolvable join dependencies, left requires {left} and right requires {right}"
        )
    base = base_candidates[0]
    final_joins.append(base)
    available_aliases.add(base.left_node)
    available_aliases.add(base.right_node)
    while final_joins_pre:
        new_final_joins_pre: List[NodeJoin] = []
        for join in final_joins_pre:
            if join.left_node in available_aliases:
                # we don't need to join twice
                # so whatever join we found first, works
                if join.right_node in available_aliases:
                    continue
                final_joins.append(join)
                available_aliases.add(join.left_node)
                available_aliases.add(join.right_node)
            else:
                new_final_joins_pre.append(join)
        final_joins_pre = new_final_joins_pre
    return final_joins



from typing import List, Optional

from trilogy.constants import logger
from trilogy.core.enums import SourceType
from trilogy.core.models.build import (
    BuildComparison,
    BuildConcept,
    BuildConditional,
    BuildDatasource,
    BuildOrderBy,
    BuildParenthetical,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.models.execute import QueryDatasource
from trilogy.core.processing.nodes.base_node import (
    StrategyNode,
    resolve_concept_map,
)
from trilogy.core.processing.utility import (
    GroupRequiredResponse,
    find_nullable_concepts,
    is_scalar_condition,
)
from trilogy.utility import unique

LOGGER_PREFIX = "[CONCEPT DETAIL - GROUP NODE]"


class GroupNode(StrategyNode):
    source_type = SourceType.GROUP

    def __init__(
        self,
        output_concepts: List[BuildConcept],
        input_concepts: List[BuildConcept],
        environment: BuildEnvironment,
        whole_grain: bool = False,
        parents: List["StrategyNode"] | None = None,
        depth: int = 0,
        partial_concepts: Optional[List[BuildConcept]] = None,
        nullable_concepts: Optional[List[BuildConcept]] = None,
        force_group: bool | None = None,
        conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        preexisting_conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        existence_concepts: List[BuildConcept] | None = None,
        hidden_concepts: set[str] | None = None,
        ordering: BuildOrderBy | None = None,
        required_outputs: List[BuildConcept] | None = None,
    ):
        super().__init__(
            input_concepts=input_concepts,
            output_concepts=output_concepts,
            environment=environment,
            whole_grain=whole_grain,
            parents=parents,
            depth=depth,
            partial_concepts=partial_concepts,
            nullable_concepts=nullable_concepts,
            force_group=force_group,
            conditions=conditions,
            existence_concepts=existence_concepts,
            preexisting_conditions=preexisting_conditions,
            hidden_concepts=hidden_concepts,
            ordering=ordering,
        )
        # the set of concepts required to preserve grain
        # set by group by node generation with aggregates
        self.required_outputs = required_outputs

    @classmethod
    def check_if_required(
        cls,
        downstream_concepts: List[BuildConcept],
        parents: list[QueryDatasource | BuildDatasource],
        environment: BuildEnvironment,
        depth: int = 0,
    ) -> GroupRequiredResponse:
        from trilogy.core.processing.discovery_utility import check_if_group_required

        return check_if_group_required(downstream_concepts, parents, environment, depth)

    def _resolve(self) -> QueryDatasource:
        parent_sources: List[QueryDatasource | BuildDatasource] = [
            p.resolve() for p in self.parents
        ]

        grains = self.check_if_required(
            self.output_concepts, parent_sources, self.environment, self.depth
        )
        target_grain = grains.target
        comp_grain = grains.upstream
        # dynamically select if we need to group
        # because sometimes, we are already at required grain
        if not grains.required and self.force_group is not True:
            # otherwise if no group by, just treat it as a select
            source_type = SourceType.SELECT
        else:
            logger.info(
                f"{self.logging_prefix}{LOGGER_PREFIX} Group node has different grain than parents; group is required."
                f" Upstream grains {[str(source.grain) for source in parent_sources]}"
                f" with final grain {comp_grain} vs"
                f" target grain {target_grain}"
                f" delta: {comp_grain - target_grain}"
            )
            source_type = SourceType.GROUP
        source_map = resolve_concept_map(
            parent_sources,
            targets=(
                unique(
                    self.output_concepts + self.conditions.concept_arguments,
                    "address",
                )
                if self.conditions
                else self.output_concepts
            ),
            inherited_inputs=self.input_concepts + self.existence_concepts,
        )
        nullable_addresses = find_nullable_concepts(
            source_map=source_map, joins=[], datasources=parent_sources
        )
        nullable_concepts = [
            x for x in self.output_concepts if x.address in nullable_addresses
        ]
        base = QueryDatasource(
            input_concepts=self.input_concepts,
            output_concepts=self.output_concepts,
            datasources=parent_sources,
            source_type=source_type,
            source_map=source_map,
            joins=[],
            grain=target_grain,
            partial_concepts=self.partial_concepts,
            nullable_concepts=nullable_concepts,
            hidden_concepts=self.hidden_concepts,
            condition=self.conditions,
            ordering=self.ordering,
        )
        # if there is a condition on a group node and it's not scalar
        # inject an additional CTE
        if self.conditions and not is_scalar_condition(self.conditions):
            base.condition = None
            base.output_concepts = unique(
                list(base.output_concepts) + list(self.conditions.row_arguments),
                "address",
            )
            # re-visible any hidden concepts
            base.hidden_concepts = set(
                [x for x in base.hidden_concepts if x not in base.output_concepts]
            )
            source_map = resolve_concept_map(
                [base],
                targets=self.output_concepts,
                inherited_inputs=base.output_concepts,
            )
            return QueryDatasource(
                input_concepts=base.output_concepts,
                output_concepts=self.output_concepts,
                datasources=[base],
                source_type=SourceType.SELECT,
                source_map=source_map,
                joins=[],
                grain=target_grain,
                nullable_concepts=base.nullable_concepts,
                partial_concepts=self.partial_concepts,
                condition=self.conditions,
                hidden_concepts=self.hidden_concepts,
                ordering=self.ordering,
            )
        return base

    def copy(self) -> "GroupNode":
        return GroupNode(
            input_concepts=list(self.input_concepts),
            output_concepts=list(self.output_concepts),
            environment=self.environment,
            whole_grain=self.whole_grain,
            parents=self.parents,
            depth=self.depth,
            partial_concepts=list(self.partial_concepts),
            nullable_concepts=list(self.nullable_concepts),
            force_group=self.force_group,
            conditions=self.conditions,
            preexisting_conditions=self.preexisting_conditions,
            existence_concepts=list(self.existence_concepts),
            hidden_concepts=set(self.hidden_concepts),
            ordering=self.ordering,
            required_outputs=(
                list(self.required_outputs) if self.required_outputs else None
            ),
        )



from os import environ
from typing import List, Optional

from trilogy.ai.enums import Provider
from trilogy.ai.models import LLMMessage, LLMResponse, UsageDict

from .base import LLMProvider, LLMRequestOptions
from .utils import RetryOptions, fetch_with_retry


class OpenAIProvider(LLMProvider):
    def __init__(
        self,
        name: str,
        model: str,
        api_key: str | None = None,
        retry_options: Optional[RetryOptions] = None,
    ):
        api_key = api_key or environ.get("OPENAI_API_KEY")
        if not api_key:
            raise ValueError(
                "API key argument or environment variable OPENAI_API_KEY is required"
            )
        super().__init__(name, api_key, model, Provider.OPENAI)
        self.base_completion_url = "https://api.openai.com/v1/chat/completions"
        self.base_model_url = "https://api.openai.com/v1/models"
        self.models: List[str] = []
        self.type = Provider.OPENAI

        self.retry_options = retry_options or RetryOptions(
            max_retries=3,
            initial_delay_ms=1000,
            retry_status_codes=[429, 500, 502, 503, 504],  # Add common API error codes
            on_retry=lambda attempt, delay_ms, error: print(
                f"Retry attempt {attempt} after {delay_ms}ms delay due to error: {str(error)}"
            ),
        )

    def generate_completion(
        self, options: LLMRequestOptions, history: List[LLMMessage]
    ) -> LLMResponse:
        try:
            import httpx
        except ImportError:
            raise ImportError(
                "Missing httpx. Install pytrilogy[ai] to use OpenAIProvider."
            )

        messages: List[dict] = []
        messages = [{"role": msg.role, "content": msg.content} for msg in history]
        try:

            def make_request():
                with httpx.Client(timeout=30) as client:
                    payload = {
                        "model": self.model,
                        "messages": messages,
                    }

                    response = client.post(
                        url=self.base_completion_url,
                        headers={
                            "Content-Type": "application/json",
                            "Authorization": f"Bearer {self.api_key}",
                        },
                        json=payload,
                    )
                    response.raise_for_status()
                    return response.json()

            data = fetch_with_retry(make_request, self.retry_options)
            return LLMResponse(
                text=data["choices"][0]["message"]["content"],
                usage=UsageDict(
                    prompt_tokens=data["usage"]["prompt_tokens"],
                    completion_tokens=data["usage"]["completion_tokens"],
                    total_tokens=data["usage"]["total_tokens"],
                ),
            )
        except httpx.HTTPStatusError as error:
            # Capture the response body text
            error_detail = error.response.text
            raise Exception(
                f"OpenAI API error ({error.response.status_code}): {error_detail}"
            )

        except Exception as error:
            raise Exception(f"OpenAI API error: {str(error)}")



# from trilogy.compiler import compile
from trilogy import Dialects
from trilogy.core.models.author import Grain
from trilogy.core.models.environment import Environment
from trilogy.core.query_processor import process_query
from trilogy.core.statements.author import SelectStatement
from trilogy.dialect.bigquery import BigqueryDialect
from trilogy.parser import parse


def test_select():
    declarations = """
key user_id int metadata(description="the description");
property user_id.display_name string metadata(description="The display name ");
property user_id.about_me string metadata(description="User provided description");
key post_id int;


datasource posts (
    user_id: user_id,
    id: post_id
    )
    grain (post_id)
    address `bigquery-public-data.stackoverflow.post_history`
;


datasource users (
    id: user_id,
    display_name: display_name,
    about_me: about_me,
    )
    grain (user_id)
    address `bigquery-public-data.stackoverflow.users`
;

    auto post_count <- count(post_id);

    """
    env, parsed = parse(declarations)

    q1 = """select
    # a comment
    user_id,
    about_me,
    post_count
;"""
    env, parse_one = parse(q1, environment=env)

    select: SelectStatement = parse_one[-1]
    assert select.grain == Grain(components=[env.concepts["user_id"]])

    q2 = """select
    about_me,
    post_count
;"""
    env, parse_two = parse(q2, environment=env)

    select: SelectStatement = parse_two[-1]
    assert (
        select.grain.components
        == Grain(components=[env.concepts["about_me"]]).components
    )

    assert (
        str(select)
        == """SELECT
    about_me,
    post_count,
;"""
    )


def test_double_aggregate():
    declarations = """
    key user_id int metadata(description="the description");
    property user_id.display_name string metadata(description="The display name ");
    property user_id.about_me string metadata(description="User provided description");
    key post_id int;


    datasource posts (
        user_id: user_id,
        id: post_id
        )
        grain (post_id)
        address `bigquery-public-data.stackoverflow.post_history`
    ;


    datasource users (
        id: user_id,
        display_name: display_name,
        about_me: about_me,
        )
        grain (user_id)
        address `bigquery-public-data.stackoverflow.users`
    ;


        """
    env, parsed = parse(declarations)

    q1 = """
    metric post_count<- count(post_id);
    metric distinct_post_count <- count_distinct(post_id);
    
    metric user_count <- count(user_id);
    
    select
        post_count,
        distinct_post_count,
        user_count
    ;"""
    env, parsed = parse(q1, environment=env)
    select: SelectStatement = parsed[-1]

    query = process_query(statement=select, environment=env)

    generator = BigqueryDialect()
    generator.compile_statement(query)


def test_modifiers():

    q1 = """
    const a <- 1;
    const b <- 2;
    
    select
        a,
        --b,
    where b =2
    ;"""
    env, parsed = parse(q1)
    select: SelectStatement = parsed[-1]
    assert select.hidden_components == set([env.concepts["b"].address])
    assert select.output_components == [env.concepts["a"], env.concepts["b"]]
    query = process_query(statement=select, environment=env)

    generator = BigqueryDialect()

    text = generator.compile_statement(query)
    assert ":b = 2" in text
    assert "as `b`" not in text


def test_having_without_select():
    q1 = """
    const a <- 1;
    const b <- 2;
    
    select
        a,

    having b =2
    ;"""
    failed = False
    try:
        env, parsed = parse(q1)
    except Exception as e:
        assert "that is not in the select projection in the HAVING clause" in str(e)
        failed = True
    assert failed


def test_local_select_concepts():
    q1 = """

key id int;
datasource local (

    id: id
    )
    grain (id)
query '''

SELECT 1 as id
''';


select id + 2 as three;

"""

    env, parsed = parse(q1)

    result = Dialects.DUCK_DB.default_executor(environment=env).execute_text(q1)[-1]
    assert result.fetchone().three == 3


def test_select_from_components():
    env = Environment()
    q1 = """

key id int;
property id.class int;
property id.name string;

select
    class,
    upper(name)-> upper_name,
    count(id) ->class_id_count,
;
"""
    env, statements = env.parse(q1)

    select: SelectStatement = statements[-1]

    assert select.grain.components == {"local.class", "local.upper_name"}
    assert select.local_concepts["local.class_id_count"].grain.components == {
        "local.class",
        "local.upper_name",
    }


def test_select_nested_agg_grain():
    env = Environment()
    q1 = """

key id int;
property id.class int;
property id.name string;

auto class_count <- count(class);
auto name_count <- count(name) by class;

auto test_ratio <- name_count / sum name_count over class;

select
    class,
    count(class) -> name_count,
    name_count / sum name_count over class -> name_class_ratio,
;
"""
    env, statements = env.parse(q1)

    select: SelectStatement = statements[-1]
    assert select.grain.components == {
        "local.class",
    }
    benv = env.materialize_for_select()
    assert benv.concepts["test_ratio"].keys == {"local.class", "local.id"}



from collections import defaultdict
from pathlib import Path
from typing import Any

from trilogy import Environment, Executor
from trilogy.authoring import ConceptDeclarationStatement, Datasource
from trilogy.core.exceptions import (
    DatasourceColumnBindingData,
    DatasourceColumnBindingError,
)
from trilogy.core.validation.environment import validate_environment
from trilogy.parsing.render import Renderer


def rewrite_file_with_errors(
    statements: list[Any], errors: list[DatasourceColumnBindingError]
):
    renderer = Renderer()
    output = []
    ds_error_map: dict[str, list[DatasourceColumnBindingData]] = defaultdict(list)
    concept_error_map: dict[str, list[DatasourceColumnBindingData]] = defaultdict(list)
    for error in errors:
        if isinstance(error, DatasourceColumnBindingError):
            for x in error.errors:
                if error.dataset_address not in ds_error_map:
                    ds_error_map[error.dataset_address] = []
                # this is by dataset address
                if x.is_modifier_issue():
                    ds_error_map[error.dataset_address].append(x)
                # this is by column
                if x.is_type_issue():
                    concept_error_map[x.address].append(x)
    for statement in statements:
        if isinstance(statement, Datasource):
            if statement.identifier in ds_error_map:
                error_cols = ds_error_map[statement.identifier]
                for col in statement.columns:
                    if col.concept.address in [x.address for x in error_cols]:
                        error_col = [
                            x for x in error_cols if x.address == col.concept.address
                        ][0]
                        col.modifiers = list(
                            set(col.modifiers + error_col.value_modifiers)
                        )
        elif isinstance(statement, ConceptDeclarationStatement):
            if statement.concept.address in concept_error_map:
                error_cols = concept_error_map[statement.concept.address]
                statement.concept.datatype = error_cols[0].value_type
        output.append(statement)

    return renderer.render_statement_string(output)


DEPTH_CUTOFF = 3


def validate_and_rewrite(
    input: Path | str, exec: Executor | None = None, depth: int = 0
) -> str | None:
    if depth > DEPTH_CUTOFF:
        print(f"Reached depth cutoff of {DEPTH_CUTOFF}, stopping.")
        return None
    if isinstance(input, str):
        raw = input
        env = Environment()
    else:
        with open(input, "r") as f:
            raw = f.read()
        env = Environment(working_path=input.parent)
    if exec:
        env = exec.environment
    env, statements = env.parse(raw)

    validation_results = validate_environment(env, exec=exec, generate_only=True)

    errors = [
        x.result
        for x in validation_results
        if isinstance(x.result, DatasourceColumnBindingError)
    ]

    if not errors:
        print("No validation errors found")
        return None
    print(
        f"Found {len(errors)} validation errors, attempting to fix, current depth: {depth}..."
    )
    for error in errors:
        for item in error.errors:
            print(f"- {item.format_failure()}")

    new_text = rewrite_file_with_errors(statements, errors)

    while iteration := validate_and_rewrite(new_text, exec=exec, depth=depth + 1):
        depth = depth + 1
        if depth >= DEPTH_CUTOFF:
            break
        if iteration:
            new_text = iteration
        depth += 1
    if isinstance(input, Path):
        with open(input, "w") as f:
            f.write(new_text)
        return None
    else:
        return new_text
